From 3c6d72cc3314afd792ac7b04c65157326b2bbd07 Mon Sep 17 00:00:00 2001
From: Wu Jianhua <jianhua.wu@intel.com>
Date: Tue, 11 Jan 2022 18:16:14 +0800
Subject: [PATCH 2/5] avfilter: add shader_vulkan filter

The following command is on how to apply shader_vulkan filter:

ffmpeg -init_hw_device vulkan -i input.264 -vf \
hwupload,shader_vulkan=path=Example.glsl:local_size_x=32:\
local_size_y=1:filter=nearest,hwdownload,format=yuv420p \
output.264

Example.glsl:

// vf_shader_vulkan filter example

void main()
{
    for (int i = 0; i < ff_Planes; i++) {
        const ivec2 size = imageSize(ff_OutputImages[i]);
        ivec2 pos = ff_Position;
        if (IS_WITHIN(pos, size)) {
            vec4 res = texture(ff_InputImages[i], pos / 1.5);
            imageStore(ff_OutputImages[i], pos, res);
        }
    }
}

Signed-off-by: Wu Jianhua <jianhua.wu@intel.com>
---
 configure                      |   1 +
 libavfilter/Makefile           |   1 +
 libavfilter/allfilters.c       |   1 +
 libavfilter/vf_shader_vulkan.c | 335 +++++++++++++++++++++++++++++++++
 4 files changed, 338 insertions(+)
 create mode 100644 libavfilter/vf_shader_vulkan.c

diff --git a/configure b/configure
index 94f513288a..ae288ab2f7 100755
--- a/configure
+++ b/configure
@@ -3713,6 +3713,7 @@ scale_filter_deps="swscale"
 scale_qsv_filter_deps="libmfx"
 scdet_filter_select="scene_sad"
 select_filter_select="scene_sad"
+shader_vulkan_filter_deps="vulkan spirv_compiler"
 sharpness_vaapi_filter_deps="vaapi"
 showcqt_filter_deps="avformat swscale"
 showcqt_filter_suggest="libfontconfig libfreetype"
diff --git a/libavfilter/Makefile b/libavfilter/Makefile
index 282967144b..2d6b8bb737 100644
--- a/libavfilter/Makefile
+++ b/libavfilter/Makefile
@@ -440,6 +440,7 @@ OBJS-$(CONFIG_SETPTS_FILTER)                 += setpts.o
 OBJS-$(CONFIG_SETRANGE_FILTER)               += vf_setparams.o
 OBJS-$(CONFIG_SETSAR_FILTER)                 += vf_aspect.o
 OBJS-$(CONFIG_SETTB_FILTER)                  += settb.o
+OBJS-$(CONFIG_SHADER_VULKAN_FILTER)          += vf_shader_vulkan.o vulkan.o vulkan_filter.o
 OBJS-$(CONFIG_SHARPEN_NPP_FILTER)            += vf_sharpen_npp.o
 OBJS-$(CONFIG_SHARPNESS_VAAPI_FILTER)        += vf_misc_vaapi.o vaapi_vpp.o
 OBJS-$(CONFIG_SHEAR_FILTER)                  += vf_shear.o
diff --git a/libavfilter/allfilters.c b/libavfilter/allfilters.c
index 714468afce..5f6382feb8 100644
--- a/libavfilter/allfilters.c
+++ b/libavfilter/allfilters.c
@@ -420,6 +420,7 @@ extern const AVFilter ff_vf_setpts;
 extern const AVFilter ff_vf_setrange;
 extern const AVFilter ff_vf_setsar;
 extern const AVFilter ff_vf_settb;
+extern const AVFilter ff_vf_shader_vulkan;
 extern const AVFilter ff_vf_sharpen_npp;
 extern const AVFilter ff_vf_sharpness_vaapi;
 extern const AVFilter ff_vf_shear;
diff --git a/libavfilter/vf_shader_vulkan.c b/libavfilter/vf_shader_vulkan.c
new file mode 100644
index 0000000000..55807425a6
--- /dev/null
+++ b/libavfilter/vf_shader_vulkan.c
@@ -0,0 +1,335 @@
+/*
+ * copyright (c) 2022 Wu Jianhua <jianhua.wu@intel.com>
+ * This file is part of FFmpeg.
+ *
+ * FFmpeg is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
+ *
+ * FFmpeg is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ *
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with FFmpeg; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
+ */
+
+#include "libavutil/random_seed.h"
+#include "libavutil/opt.h"
+#include "vulkan_filter.h"
+#include "internal.h"
+
+typedef struct ShaderVulkanContext {
+    FFVulkanContext vkctx;
+    FFVkQueueFamilyCtx qf;
+    FFVkExecContext *exec;
+    FFVulkanPipeline *pl;
+    VkDescriptorImageInfo input_images[3];
+    VkDescriptorImageInfo output_images[3];
+
+    char *input_name;
+    char *output_name;
+    char *filepath;
+
+    VkFilter filter;
+    int local_size_x;
+    int local_size_y;
+    int initialized;
+} ShaderVulkanContext;
+
+#define BUF_LEN 4096
+
+static int load_shader_from_file(ShaderVulkanContext *s, FFVkSPIRVShader *shd)
+{
+    int len, err = 0;
+    char src[BUF_LEN];
+    FILE *fp;
+
+    if (!s->filepath) {
+        av_log(s, AV_LOG_ERROR, "Input shader language source "
+               "file not specified yet\n");
+        return AVERROR(EINVAL);
+    }
+
+    fp = av_fopen_utf8(s->filepath, "r");
+    if (!fp) {
+        av_log(s, AV_LOG_ERROR, "Unable to open shader language "
+               "source file \"%s\".\n", s->filepath);
+        return AVERROR(ENOENT);
+    }
+
+    while (!feof(fp)) {
+        len = fread(src, 1, BUF_LEN - 1, fp);
+        if (len == 0 && ferror(fp)) {
+            err = AVERROR(EIO);
+            goto fail;
+        }
+
+        src[len] = '\0';
+        av_bprintf(&shd->src, "%s", src);
+    }
+
+fail:
+    fclose(fp);
+    return err;
+}
+
+static av_cold int init_filter(AVFilterContext *avctx, AVFrame *in)
+{
+    int err = 0;
+    FFVkSPIRVShader *shd;
+    ShaderVulkanContext *s = avctx->priv;
+
+    const int planes = av_pix_fmt_count_planes(s->vkctx.output_format);
+
+    FFVkSampler *sampler = ff_vk_init_sampler(&s->vkctx, 1, s->filter);
+    if (!sampler)
+        return AVERROR_EXTERNAL;
+
+    {
+        FFVulkanDescriptorSetBinding image_descs[] = {
+            {
+                .name       = s->input_name,
+                .type       = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER,
+                .dimensions = 2,
+                .elems      = planes,
+                .stages     = VK_SHADER_STAGE_COMPUTE_BIT,
+                .sampler    = sampler,
+                .updater    = s->input_images,
+            },
+            {
+                .name       = s->output_name,
+                .type       = VK_DESCRIPTOR_TYPE_STORAGE_IMAGE,
+                .mem_layout = ff_vk_shader_rep_fmt(s->vkctx.output_format),
+                .mem_quali  = "writeonly",
+                .dimensions = 2,
+                .elems      = planes,
+                .stages     = VK_SHADER_STAGE_COMPUTE_BIT,
+                .updater    = s->output_images,
+            },
+        };
+
+        ff_vk_qf_init(&s->vkctx, &s->qf, VK_QUEUE_COMPUTE_BIT, 0);
+
+        s->pl = ff_vk_create_pipeline(&s->vkctx, &s->qf);
+        if (!s->pl) {
+            err = AVERROR(ENOMEM);
+            goto fail;
+        }
+
+        shd = ff_vk_init_shader(s->pl, "shader_compute", image_descs[0].stages);
+        if (!shd) {
+            err = AVERROR(ENOMEM);
+            goto fail;
+        }
+
+        ff_vk_set_compute_shader_sizes(shd, (int [3]){ s->local_size_x, s->local_size_y, 1 });
+        RET(ff_vk_add_descriptor_set(&s->vkctx, s->pl, shd, image_descs, FF_ARRAY_ELEMS(image_descs), 0));
+
+        GLSLF(0, #define ff_Planes %i ,planes);
+        GLSLC(0, #define ff_Position ivec2(gl_GlobalInvocationID.xy));
+        GLSLC(0, );
+        RET(load_shader_from_file(s, shd));
+
+        RET(ff_vk_compile_shader(&s->vkctx, shd, "main"));
+        RET(ff_vk_init_pipeline_layout(&s->vkctx, s->pl));
+        RET(ff_vk_init_compute_pipeline(&s->vkctx, s->pl));
+    }
+
+    RET(ff_vk_create_exec_ctx(&s->vkctx, &s->exec, &s->qf));
+
+    s->initialized = 1;
+
+fail:
+    return err;
+}
+
+static av_cold void uninit(AVFilterContext *avctx)
+{
+    ShaderVulkanContext *s = avctx->priv;
+
+    ff_vk_uninit(&s->vkctx);
+    s->initialized = 0;
+}
+
+static int process_frames(AVFilterContext *avctx, AVFrame *out_frame, AVFrame *in_frame)
+{
+    int err;
+    VkCommandBuffer cmd_buf;
+    ShaderVulkanContext *s = avctx->priv;
+    FFVulkanFunctions *vk = &s->vkctx.vkfn;
+
+    const int planes = av_pix_fmt_count_planes(s->vkctx.output_format);
+
+    AVVkFrame *in  = (AVVkFrame *)in_frame->data[0];
+    AVVkFrame *out = (AVVkFrame *)out_frame->data[0];
+
+    const VkFormat *input_formats  = av_vkfmt_from_pixfmt(s->vkctx.input_format);
+    const VkFormat *output_formats = av_vkfmt_from_pixfmt(s->vkctx.output_format);
+
+    ff_vk_start_exec_recording(&s->vkctx, s->exec);
+    cmd_buf = ff_vk_get_exec_buf(s->exec);
+
+    for (int i = 0; i < planes; i++) {
+        RET(ff_vk_create_imageview(&s->vkctx, s->exec, &s->input_images[i].imageView,
+                                   in->img[i],
+                                   input_formats[i],
+                                   ff_comp_identity_map));
+
+        RET(ff_vk_create_imageview(&s->vkctx, s->exec, &s->output_images[i].imageView,
+                                   out->img[i],
+                                   output_formats[i],
+                                   ff_comp_identity_map));
+
+        s->input_images[i].imageLayout  = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
+        s->output_images[i].imageLayout = VK_IMAGE_LAYOUT_GENERAL;
+    }
+
+    ff_vk_update_descriptor_set(&s->vkctx, s->pl, 0);
+
+    for (int i = 0; i < planes; i++) {
+        VkImageMemoryBarrier barriers[] = {
+            {
+                .sType                       = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,
+                .srcAccessMask               = 0,
+                .dstAccessMask               = VK_ACCESS_SHADER_READ_BIT,
+                .oldLayout                   = in->layout[i],
+                .newLayout                   = s->input_images[i].imageLayout,
+                .srcQueueFamilyIndex         = VK_QUEUE_FAMILY_IGNORED,
+                .dstQueueFamilyIndex         = VK_QUEUE_FAMILY_IGNORED,
+                .image                       = in->img[i],
+                .subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
+                .subresourceRange.levelCount = 1,
+                .subresourceRange.layerCount = 1,
+            },
+            {
+                .sType                       = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER,
+                .srcAccessMask               = 0,
+                .dstAccessMask               = VK_ACCESS_SHADER_WRITE_BIT,
+                .oldLayout                   = out->layout[i],
+                .newLayout                   = s->output_images[i].imageLayout,
+                .srcQueueFamilyIndex         = VK_QUEUE_FAMILY_IGNORED,
+                .dstQueueFamilyIndex         = VK_QUEUE_FAMILY_IGNORED,
+                .image                       = out->img[i],
+                .subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT,
+                .subresourceRange.levelCount = 1,
+                .subresourceRange.layerCount = 1,
+            },
+        };
+
+        vk->CmdPipelineBarrier(cmd_buf, VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT,
+                               VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT, 0,
+                               0, NULL, 0, NULL, FF_ARRAY_ELEMS(barriers), barriers);
+
+        in->layout[i]  = barriers[0].newLayout;
+        in->access[i]  = barriers[0].dstAccessMask;
+
+        out->layout[i] = barriers[1].newLayout;
+        out->access[i] = barriers[1].dstAccessMask;
+    }
+
+    ff_vk_bind_pipeline_exec(&s->vkctx, s->exec, s->pl);
+
+    vk->CmdDispatch(cmd_buf, FFALIGN(s->vkctx.output_width, s->local_size_x) / s->local_size_x,
+                    FFALIGN(s->vkctx.output_height, s->local_size_y) / s->local_size_y, 1);
+
+    ff_vk_add_exec_dep(&s->vkctx, s->exec, in_frame, VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT);
+    ff_vk_add_exec_dep(&s->vkctx, s->exec, out_frame, VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT);
+
+    err = ff_vk_submit_exec_queue(&s->vkctx, s->exec);
+    if (err)
+        return err;
+
+    ff_vk_qf_rotate(&s->qf);
+
+    return 0;
+
+fail:
+    ff_vk_discard_exec_deps(s->exec);
+    return err;
+}
+
+static int filter_frame(AVFilterLink *link, AVFrame *in)
+{
+    int err;
+    AVFrame *out = NULL;
+    AVFilterContext *avctx = link->dst;
+    ShaderVulkanContext *s = avctx->priv;
+    AVFilterLink *outlink = avctx->outputs[0];
+
+    out = ff_get_video_buffer(outlink, outlink->w, outlink->h);
+    if (!out) {
+        err = AVERROR(ENOMEM);
+        goto fail;
+    }
+
+    if (!s->initialized)
+        RET(init_filter(avctx, in));
+
+    RET(process_frames(avctx, out, in));
+
+    RET(av_frame_copy_props(out, in));
+
+    av_frame_free(&in);
+
+    return ff_filter_frame(outlink, out);
+
+fail:
+    av_frame_free(&in);
+    av_frame_free(&out);
+
+    return err;
+}
+
+#define OFFSET(x) offsetof(ShaderVulkanContext, x)
+#define FLAGS (AV_OPT_FLAG_FILTERING_PARAM | AV_OPT_FLAG_VIDEO_PARAM)
+
+static const AVOption shader_vulkan_options[] = {
+    { "path",   "Path to the shader file",       OFFSET(filepath),    AV_OPT_TYPE_STRING, { .str = NULL              }, 0, 0, FLAGS },
+    { "input",  "Define name for input images",  OFFSET(input_name),  AV_OPT_TYPE_STRING, { .str = "ff_InputImages"  }, 0, 0, FLAGS },
+    { "output", "Define name for output images", OFFSET(output_name), AV_OPT_TYPE_STRING, { .str = "ff_OutputImages" }, 0, 0, FLAGS },
+
+    { "filter", "Set filter for sampler", OFFSET(filter), AV_OPT_TYPE_INT, { .i64 = VK_FILTER_LINEAR }, 0, VK_FILTER_LINEAR, FLAGS, "filter" },
+        { "linear",  "Filter frame in linear interpolation",  0, AV_OPT_TYPE_CONST, { .i64 = VK_FILTER_LINEAR  }, 0, 0, .flags = FLAGS, "filter" },
+        { "nearest", "Filter frame in nearest interpolation", 0, AV_OPT_TYPE_CONST, { .i64 = VK_FILTER_NEAREST }, 0, 0, .flags = FLAGS, "filter" },
+
+    { "local_size_x", "Set compute local size x", OFFSET(local_size_x), AV_OPT_TYPE_INT, { .i64 = 32 }, 0, 128, FLAGS },
+    { "local_size_y", "Set compute local size y", OFFSET(local_size_y), AV_OPT_TYPE_INT, { .i64 = 32 }, 0, 128, FLAGS },
+
+    { NULL },
+};
+
+AVFILTER_DEFINE_CLASS(shader_vulkan);
+
+static const AVFilterPad shader_vulkan_inputs[] = {
+    {
+        .name         = "default",
+        .type         = AVMEDIA_TYPE_VIDEO,
+        .filter_frame = &filter_frame,
+        .config_props = &ff_vk_filter_config_input,
+    }
+};
+
+static const AVFilterPad shader_vulkan_outputs[] = {
+    {
+        .name         = "default",
+        .type         = AVMEDIA_TYPE_VIDEO,
+        .config_props = &ff_vk_filter_config_output,
+    }
+};
+
+const AVFilter ff_vf_shader_vulkan = {
+    .name           = "shader_vulkan",
+    .description    = NULL_IF_CONFIG_SMALL("Filter frame in shader language"),
+    .priv_size      = sizeof(ShaderVulkanContext),
+    .init           = &ff_vk_filter_init,
+    .uninit         = &uninit,
+    FILTER_INPUTS(shader_vulkan_inputs),
+    FILTER_OUTPUTS(shader_vulkan_outputs),
+    FILTER_SINGLE_PIXFMT(AV_PIX_FMT_VULKAN),
+    .priv_class     = &shader_vulkan_class,
+    .flags_internal = FF_FILTER_FLAG_HWFRAME_AWARE,
+};
-- 
2.25.1

